"""Define S3Uploader class."""
import asyncio
import os
import posixpath
import threading
from http import HTTPStatus
from typing import List, Optional

import aioboto3
import boto3
from botocore.awsrequest import AWSRequest, AWSResponse
from tqdm.autonotebook import tqdm

from ML_management.mlmanagement import get_s3_gateway_url, variables
from ML_management.mlmanagement.mlmanager import tar_folder
from ML_management.mlmanagement.session import AuthSession
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.uploader_data.utils import get_upload_paths

MAX_TASKS_NUMBER = 50


class S3Uploader:
    """S3 uploader files class."""

    def __init__(self):
        """Init creds."""
        self.default_url = get_s3_gateway_url()
        self.default_access_key_id, self.default_secret_access_key = variables._get_s3_credentials()
        self.session = AuthSession(gateway_url=self.default_url)

    def upload(
        self,
        local_path: str,
        bucket: str,
        upload_as_tar: bool = False,
        new_bucket_visibility: VisibilityOptions = VisibilityOptions.PRIVATE,
        verbose: bool = True,
    ) -> Optional[asyncio.Task]:
        """
        Upload directory to bucket.

        Parameters
        ----------
        local_path: str
            path to directory with files you want to upload.
        bucket: str
            name of bucket you want to upload to.
        upload_as_tar: bool = False
            If the option is set to True, the files will be uploaded as a single tar archive. Default: False
        verbose: bool = True
            If the option is set to True and upload_as_tar set to False,
            a progress bar with the number of uploaded files will be displayed.

        Returns
        -------
        Optional[asyncio.Task].
            If the files uploading to the bucket is started inside an asynchronous application,
            the method will schedule the task in the running event loop and
            return instance of asyncio.Task for further process monitoring by the application
        """
        if upload_as_tar:
            self._upload_as_tar(local_path=local_path, bucket=bucket, new_bucket_visibility=new_bucket_visibility)
            return

        # in case of asyncio.run() was called previously in code, it closed event loop
        # and get_event_loop() will raise exception:
        # RuntimeError: There is no current event loop in thread <THREAD_NAME>
        try:
            if asyncio.get_event_loop().is_running():
                return asyncio.create_task(
                    self._async_upload_files(
                        local_path=local_path,
                        bucket=bucket,
                        new_bucket_visibility=new_bucket_visibility,
                        verbose=verbose,
                    )
                )
            asyncio.get_event_loop().close()
        except Exception:
            pass
        asyncio.run(
            self._async_upload_files(
                local_path=local_path, bucket=bucket, new_bucket_visibility=new_bucket_visibility, verbose=verbose
            )
        )

    def _upload_as_tar(self, local_path: str, bucket: str, new_bucket_visibility: VisibilityOptions) -> None:
        local_path = os.path.normpath(local_path)
        if not os.path.exists(local_path):
            raise FileNotFoundError(f"Path: {local_path} does not exist")

        r, w = os.pipe()

        try:
            thread = threading.Thread(target=tar_folder, args=(w, local_path), daemon=True)
            thread.start()
        except Exception as err:
            os.close(r)
            os.close(w)
            raise err

        s3_client = boto3.client(
            service_name="s3",
            use_ssl=True,
            endpoint_url=posixpath.join(self.default_url, "s3/"),
            aws_access_key_id=self.default_access_key_id,
            aws_secret_access_key=self.default_secret_access_key,
        )
        event_system = s3_client.meta.events
        event_system.register("before-sign.s3.*", self._add_auth_cookies)
        event_system.register("after-call.s3.*", self._update_auth_cookies)
        buckets = self._list_buckets()
        if bucket not in buckets:
            self._create_bucket(name=bucket, visibility=new_bucket_visibility)
        try:
            with open(r, "rb") as fileobj:
                s3_client.upload_fileobj(Fileobj=fileobj, Bucket=bucket, Key=f"{os.path.basename(local_path)}.tar")
        finally:
            thread.join()

    async def _async_upload_files(
        self, local_path: str, bucket: str, new_bucket_visibility: VisibilityOptions, verbose: bool = True
    ):
        local_path = os.path.normpath(local_path)
        if not os.path.exists(local_path):
            raise FileNotFoundError(f"Path: {local_path} does not exist")

        session = aioboto3.Session(
            aws_access_key_id=self.default_access_key_id, aws_secret_access_key=self.default_secret_access_key
        )
        async with session.client(
            "s3", use_ssl=True, endpoint_url=posixpath.join(self.default_url, "s3/")
        ) as s3_client:
            event_system = s3_client.meta.events
            event_system.register("before-sign.s3.*", self._add_auth_cookies)
            event_system.register("after-call.s3.*", self._update_auth_cookies)
            buckets = self._list_buckets()
            if bucket not in buckets:
                self._create_bucket(name=bucket, visibility=new_bucket_visibility)

            upload_paths = get_upload_paths(local_path)
            total_tasks = len(upload_paths)

            with tqdm(
                total=total_tasks,
                disable=not verbose,
                unit="Files",
                unit_scale=True,
                unit_divisor=1024,
            ) as pbar:
                current_task_num = 0
                while current_task_num < total_tasks:
                    tasks: list[asyncio.Task] = []
                    for path in upload_paths[current_task_num : current_task_num + MAX_TASKS_NUMBER]:
                        tasks.append(
                            asyncio.create_task(
                                s3_client.upload_file(Filename=path.local_path, Bucket=bucket, Key=path.storage_path)
                            )
                        )
                    await asyncio.gather(*tasks)
                    pbar.update(len(tasks))
                    current_task_num += MAX_TASKS_NUMBER

    # arguments to callback are passed like kwargs, so kwargs must be present in signature
    def _add_auth_cookies(self, request: AWSRequest, **kwargs) -> None:  # noqa
        request.headers.add_header("Cookie", self.session._get_cookie_header())

    # arguments to callback are passed like kwargs, so kwargs must be present in signature
    def _update_auth_cookies(self, http_response: AWSResponse, **kwargs) -> None:  # noqa
        cookie_header = http_response.headers.get("set-cookie")
        if cookie_header is None:
            return
        cookies: list[str] = cookie_header.split("; ")
        for cookie in cookies:
            if "kc-access" not in cookie:
                continue
            _, new_access_token = cookie.split("=", maxsplit=1)
            self.session.cookies["kc-access"] = new_access_token
            break

    def _list_buckets(self) -> List[str]:
        with self.session.get(posixpath.join(self.default_url, "list-buckets")) as response:
            return response.json()["buckets"]

    def _create_bucket(self, name: str, visibility: VisibilityOptions) -> None:
        with self.session.post(
            posixpath.join(self.default_url, "create-bucket"), json={"name": name, "visibility": visibility.value}
        ) as response:
            if response.status_code != HTTPStatus.CREATED:
                raise RuntimeError(f"Failed to create bucket: {response.text}")



"""Module with patterns for models parameters in executors."""
from typing import List, Tuple

from pydantic import BaseModel

from ML_management.dataset_loader.dataset_loader_pattern_to_methods_map import DatasetLoaderMethodName
from ML_management.executor.upload_model_mode import UploadModelMode
from ML_management.model.model_type_to_methods_map import ModelMethodName

DEFAULT_ROLE = "single"


class OneModelPattern(BaseModel):
    """Pattern for only one model."""

    upload_model_mode: UploadModelMode
    desired_model_methods: List[ModelMethodName]

    def serialize(self) -> Tuple[dict, dict]:
        return {DEFAULT_ROLE: self.upload_model_mode}, {DEFAULT_ROLE: self.desired_model_methods}


class OneModelPatternWithRole(OneModelPattern):
    """Pattern for only one model with specific role."""

    role: str

    def serialize(self) -> Tuple[dict, dict]:
        return {self.role: self.upload_model_mode}, {self.role: self.desired_model_methods}


class OneDatasetLoaderPattern(BaseModel):
    """Pattern for only one dataset loader."""

    desired_dataset_loader_methods: List[DatasetLoaderMethodName] = [DatasetLoaderMethodName.get_dataset]

    def serialize(self) -> dict:
        return {DEFAULT_ROLE: self.desired_dataset_loader_methods}


class OneDatasetLoaderPatternWithRole(OneDatasetLoaderPattern):
    """Pattern for only one dataset loader with specific role."""

    role: str

    def serialize(self) -> dict:
        return {self.role: self.desired_dataset_loader_methods}


class ArbitraryDatasetLoaderPattern(BaseModel):
    """Pattern for arbitrary number of dataset loaders."""

    desired_dataset_loaders: List[OneDatasetLoaderPatternWithRole]

    def serialize(self) -> dict:
        result = {}
        roles_set = set()
        multiple_roles = set()
        for desired_dataset_loader in self.desired_dataset_loaders:
            if desired_dataset_loader.role in roles_set:
                multiple_roles.add(desired_dataset_loader.role)
            roles_set.add(desired_dataset_loader.role)
            result.update(desired_dataset_loader.serialize())
        if len(multiple_roles) != 0:
            raise RuntimeError(
                "Your data loaders must have different roles. "
                f"Roles {', '.join(multiple_roles)} occur 2 or more times."
            )
        return result


class ArbitraryModelsPattern(BaseModel):
    """Pattern for arbitrary number of models."""

    desired_models: List[OneModelPatternWithRole]

    def serialize(self) -> Tuple[dict, dict]:
        upload_model_modes, desired_model_methods = {}, {}
        roles_set = set()
        multiple_roles = set()
        for desired_model in self.desired_models:
            if desired_model.role in roles_set:
                multiple_roles.add(desired_model.role)
            roles_set.add(desired_model.role)
            current_upload_model_modes, current_desired_model_methods = desired_model.serialize()
            upload_model_modes.update(current_upload_model_modes)
            desired_model_methods.update(current_desired_model_methods)
        if len(multiple_roles) != 0:
            raise RuntimeError(
                f"Your models must have different roles. Roles {', '.join(multiple_roles)} occur 2 ore more times."
            )
        return upload_model_modes, desired_model_methods



from .base_executor import BaseExecutor
from .executor_pattern import JobExecutorPattern
from .no_model_executor_pattern import NoModelExecutorPattern

NoModelExecutorPattern = NoModelExecutorPattern
BaseExecutor = BaseExecutor
JobExecutorPattern = JobExecutorPattern



"""S3 Collector for downloading files and folders."""
import asyncio
import os
import posixpath
import threading
from typing import List, Optional, Union

import aioboto3
import boto3
from aiofiles import os as aos
from botocore.awsrequest import AWSRequest, AWSResponse
from botocore.exceptions import ClientError
from tqdm.autonotebook import tqdm

from ML_management.collectors.collector_pattern import CollectorPattern
from ML_management.mlmanagement import get_s3_gateway_url, variables
from ML_management.mlmanagement.mlmanager import untar_folder
from ML_management.mlmanagement.session import AuthSession

MAX_OBJECTS_PER_PAGE = 50


class S3BucketNotFoundError(Exception):
    """Define Bucket Not Found Exception."""

    def __init__(self, bucket: str):
        self.bucket = bucket
        self.message = f'Bucket "{bucket}" does not exist'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3BucketNotFoundError, (self.bucket,))


class S3ObjectNotFoundError(Exception):
    """Define Version Not Found Exception."""

    def __init__(self, path: str, bucket: str):
        self.path = path
        self.bucket = bucket
        self.message = f'Object "{path}" is not found in "{bucket}" bucket'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3ObjectNotFoundError, (self.path, self.bucket))


class FileIsNotTarError(Exception):
    """Define exception when tar is expected but not given."""

    def __init__(self, filename: str):
        self.filename = filename
        self.message = f"Expected tar file, but file {self.filename} does not have '.tar' extension"
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (FileIsNotTarError, (self.filename,))


class AmbiguousFileChoiceError(Exception):
    """Define exception when there is not exactly one file in bucket to choice."""

    def __init__(self, number_of_files: int):
        self.number_of_files = number_of_files
        if not number_of_files:
            appendix = " bucket is empty"
        else:
            appendix = "there is more than one file in bucket. Specify one file name"
        self.message = f"Expected one tar file in bucket, but {appendix}."
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (AmbiguousFileChoiceError, (self.number_of_files,))


class S3Collector(CollectorPattern):
    """Collector for S3 paths using boto3 library."""

    def __init__(self) -> None:
        """Init creds."""
        self.default_url = get_s3_gateway_url()
        self.default_access_key_id, self.default_secret_access_key = variables._get_s3_credentials()
        self.session = AuthSession(gateway_url=self.default_url)

    @staticmethod
    def get_json_schema():
        """Return json schema."""
        return {
            "type": "object",
            "properties": {
                "bucket": {"type": "string"},
                "untar_data": {"type": "boolean"},
                "remote_paths": {"type": "array", "items": {"type": "string"}},
            },
            "required": ["bucket"],
            "additionalProperties": False,
        }

    def set_data(
        self,
        *,
        local_path: str = "s3_data",
        bucket: str,  # TODO do we pass bucket or not? is it always possible to parse?
        untar_data: bool = False,
        remote_paths: Optional[List[str]] = None,
        service_name: str = "s3",
        region_name: Optional[str] = None,
        api_version: Optional[str] = None,
        use_ssl: bool = True,
        verify: Optional[Union[bool, str]] = None,
        endpoint_url: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        verbose: bool = True,
    ) -> Union[asyncio.Task, str]:
        """
        Set data.

        :type local_path: string
        :param local_path: Local path to save data to.  Defaults to /s3_data/.

        :type bucket: string
        :param bucket: Bucket containing requested files.

        :type remote_paths: list(string)
        :param remote_paths: List of paths relative to passed bucket.  Each path
            can represent either a single file, or a folder.  If a path represents
            a folder (should end with a slash), then all contents of a folder are recursively downloaded.

        :type service_name: string
        :param service_name: The name of a service, e.g. 's3' or 'ec2'
        available to boto3.  Defaults to 's3'.

        :type region_name: string
        :param region_name: The name of the region associated with the client.
            A client is associated with a single region.

        :type api_version: string
        :param api_version: The API version to use.  By default, botocore will
            use the latest API version when creating a client.  You only need
            to specify this parameter if you want to use a previous API version
            of the client.

        :type use_ssl: boolean
        :param use_ssl: Whether to use SSL.  By default, SSL is used.
            Note that not all services support non-ssl connections.

        :type verify: boolean/string
        :param verify: Whether to verify SSL certificates.  By default,
            SSL certificates are verified.  You can provide the following
            values:

            * False - do not validate SSL certificates.  SSL will still be
              used (unless use_ssl is False), but SSL certificates
              will not be verified.
            * path/to/cert/bundle.pem - A filename of the CA cert bundle to
              uses.  You can specify this argument if you want to use a
              different CA cert bundle than the one used by botocore.

        :type endpoint_url: string
        :param endpoint_url: The complete URL to use for the constructed
            client. Normally, botocore will automatically construct the
            appropriate URL to use when communicating with a service.  You
            can specify a complete URL (including the "http/https" scheme)
            to override this behavior.  If this value is provided,
            then ``use_ssl`` is ignored.

        :type aws_access_key_id: string
        :param aws_access_key_id: The access key to use when creating
            the client.  This is entirely optional, and if not provided,
            the credentials configured for the session will automatically
            be used.  You only need to provide this argument if you want
            to override the credentials used for this specific client.

        :type aws_secret_access_key: string
        :param aws_secret_access_key: The secret key to use when creating
            the client.  Same semantics as aws_access_key_id above.

        :type aws_session_token: string
        :param aws_session_token: The session token to use when creating
            the client.  Same semantics as aws_access_key_id above.

        :type verbose: bool
        :param verbose: Whether to disable the entire progressbar wrapper.
        """
        if untar_data:
            return self._download_data_tar(
                local_path=local_path,
                bucket=bucket,
                remote_paths=remote_paths,
                service_name=service_name,
                region_name=region_name,
                api_version=api_version,
                use_ssl=use_ssl,
                verify=verify,
                endpoint_url=endpoint_url,
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
            )

        try:
            if asyncio.get_event_loop().is_running():
                return asyncio.create_task(
                    self._download_files(
                        local_path=local_path,
                        bucket=bucket,
                        remote_paths=remote_paths,
                        service_name=service_name,
                        region_name=region_name,
                        api_version=api_version,
                        use_ssl=use_ssl,
                        verify=verify,
                        endpoint_url=endpoint_url,
                        aws_access_key_id=aws_access_key_id,
                        aws_secret_access_key=aws_secret_access_key,
                        aws_session_token=aws_session_token,
                        verbose=verbose,
                    )
                )
            asyncio.get_event_loop().close()
        except Exception:
            pass
        return asyncio.run(
            self._download_files(
                local_path=local_path,
                bucket=bucket,
                remote_paths=remote_paths,
                service_name=service_name,
                region_name=region_name,
                api_version=api_version,
                use_ssl=use_ssl,
                verify=verify,
                endpoint_url=endpoint_url,
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                verbose=verbose,
            )
        )

    def _download_data_tar(
        self,
        *,
        local_path: str = "s3_data",
        bucket: str,
        remote_paths: Optional[List[str]] = None,
        service_name: str = "s3",
        region_name: Optional[str] = None,
        api_version: Optional[str] = None,
        use_ssl: bool = True,
        verify: Optional[Union[bool, str]] = None,
        endpoint_url: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
    ) -> str:
        if remote_paths is not None and len(remote_paths) != 1:
            raise RuntimeError(f"Expected one tar object, but {len(remote_paths)} were given.")

        s3_client = boto3.client(
            service_name=service_name,
            region_name=region_name,
            api_version=api_version,
            use_ssl=use_ssl,
            verify=verify,
            endpoint_url=endpoint_url if endpoint_url else posixpath.join(self.default_url, "s3/"),
            aws_access_key_id=aws_access_key_id if aws_access_key_id else self.default_access_key_id,
            aws_secret_access_key=aws_secret_access_key if aws_secret_access_key else self.default_secret_access_key,
            aws_session_token=aws_session_token,
        )
        event_system = s3_client.meta.events
        event_system.register("before-sign.s3.*", self._add_auth_cookies)
        event_system.register("after-call.s3.*", self._update_auth_cookies)
        if remote_paths is not None:
            tar_name = remote_paths[0]
            if not tar_name.endswith(".tar"):
                raise FileIsNotTarError("Expected tar file, but file does not have '.tar' extension")
            return self._download_tar(
                local_path=local_path, s3_client=s3_client, tar_name=remote_paths[0], bucket=bucket
            )

        try:
            # check that there is only one file in bucket
            # If there is more than 1 file, raise exception about ambiguous choice
            paginator = s3_client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=bucket, Prefix="", PaginationConfig={"MaxItems": 2})
            first_page = next(iter(page_iterator))
            files = first_page.get("Contents", [])
            if len(files) != 1:
                raise AmbiguousFileChoiceError(number_of_files=len(files))
            tar_name: str = files[0]["Key"]
            if not tar_name.endswith(".tar"):
                raise FileIsNotTarError("Expected tar file, but file does not have '.tar' extension")
        except ClientError as err:
            if err.response["Error"]["Code"] == "NoSuchBucket":
                raise S3BucketNotFoundError(bucket=bucket) from None
            else:
                raise

        return self._download_tar(local_path=local_path, s3_client=s3_client, tar_name=tar_name, bucket=bucket)

    def _download_tar(self, local_path: str, tar_name: str, bucket: str, s3_client) -> str:
        r, w = os.pipe()
        try:
            thread = threading.Thread(target=untar_folder, args=(r, local_path), daemon=True)
            thread.start()
        except Exception as err:
            os.close(r)
            os.close(w)
            raise err

        try:
            with open(w, "wb") as fileobj:
                s3_client.download_fileobj(Fileobj=fileobj, Bucket=bucket, Key=tar_name)
        except ClientError as err:
            if err.response["Error"]["Code"] == "404":
                raise S3ObjectNotFoundError(path=tar_name, bucket=bucket) from None
            raise
        finally:
            thread.join()

        return os.path.join(local_path, tar_name[:-4])

    async def _download_files(
        self,
        *,
        local_path: str = "s3_data",
        bucket: str,
        remote_paths: Optional[List[str]] = None,
        service_name: str = "s3",
        region_name: Optional[str] = None,
        api_version: Optional[str] = None,
        use_ssl: bool = True,
        verify: Optional[Union[bool, str]] = None,
        endpoint_url: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        verbose: bool = True,
    ) -> str:
        async def get_objects_number(paginator, remote_path: str = "") -> int:
            total_objects = 0

            page_iterator = paginator.paginate(
                Bucket=bucket,
                Prefix=remote_path,
            )
            async for page in page_iterator:
                if page["KeyCount"] == 0:
                    raise S3ObjectNotFoundError(path=remote_path, bucket=bucket)
                total_objects += page["KeyCount"]

            return total_objects

        remote_paths = remote_paths if remote_paths else [""]
        session = aioboto3.Session(
            aws_access_key_id=aws_access_key_id if aws_access_key_id else self.default_access_key_id,
            aws_secret_access_key=aws_secret_access_key if aws_secret_access_key else self.default_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region_name,
        )
        async with session.client(
            service_name=service_name,
            use_ssl=use_ssl,
            verify=verify,
            api_version=api_version,
            endpoint_url=endpoint_url if endpoint_url else posixpath.join(self.default_url, "s3/"),
        ) as s3_client:
            event_system = s3_client.meta.events
            event_system.register("before-sign.s3.*", self._add_auth_cookies)
            event_system.register("after-call.s3.*", self._update_auth_cookies)
            total_tasks = 0
            try:
                paginator = s3_client.get_paginator("list_objects_v2")
                for remote_path in remote_paths:
                    total_tasks += await get_objects_number(paginator, remote_path)
            except ClientError as err:
                if err.response["Error"]["Code"] == "NoSuchBucket":
                    raise S3BucketNotFoundError(bucket=bucket) from None
                else:
                    raise

            with tqdm(
                total=total_tasks,
                unit="Files",
                unit_scale=True,
                unit_divisor=1024,
                disable=not verbose,
            ) as pbar:
                for remote_path in remote_paths:
                    page_iterator = paginator.paginate(
                        Bucket=bucket, Prefix=remote_path, PaginationConfig={"MaxItems": MAX_OBJECTS_PER_PAGE}
                    )
                    async for page in page_iterator:
                        tasks: list[asyncio.Task] = []
                        for obj in page.get("Contents", []):
                            file_path = obj.get("Key")

                            local_dir_path = os.path.join(local_path, posixpath.dirname(file_path))
                            local_file_path = os.path.join(local_path, file_path)
                            tasks.append(
                                asyncio.create_task(
                                    self._download_one_object(
                                        bucket=bucket,
                                        key=file_path,
                                        local_dir_path=local_dir_path,
                                        local_file_path=local_file_path,
                                        s3_client=s3_client,
                                    )
                                )
                            )
                        await asyncio.gather(*tasks)
                        pbar.update(len(tasks))

        return local_path

    # arguments to callback are passed like kwargs, so kwargs must be present in signature
    def _add_auth_cookies(self, request: AWSRequest, **kwargs) -> None:  # noqa
        request.headers.add_header("Cookie", self.session._get_cookie_header())

    # arguments to callback are passed like kwargs, so kwargs must be present in signature
    def _update_auth_cookies(self, http_response: AWSResponse, **kwargs) -> None:  # noqa
        cookie_header = http_response.headers.get("set-cookie")
        if cookie_header is None:
            return
        cookies: list[str] = cookie_header.split("; ")
        for cookie in cookies:
            if "kc-access" not in cookie:
                continue
            _, new_access_token = cookie.split("=", maxsplit=1)
            self.session.cookies["kc-access"] = new_access_token
            break

    async def _download_one_object(self, bucket: str, key: str, local_dir_path: str, local_file_path: str, s3_client):
        if not await aos.path.exists(local_dir_path):
            await aos.makedirs(local_dir_path, exist_ok=True)
        await s3_client.download_file(Bucket=bucket, Key=key, Filename=local_file_path)



"""General testing of the sdk."""
import os
import unittest

import pandas

from ML_management.mlmanagement import set_server_url
from ML_management.sdk import sdk
from ML_management.sdk.parameters import (
    ArbitraryModels,
    DatasetLoaderMethodParams,
    ModelMethodParams,
    ModelVersionChoice,
    ModelWithRole,
    SingleDatasetLoader,
)
from ML_management.sdk.schema import ExecutionJob, ModelVersion

os.environ["kc_access"] = "kc_access"
os.environ["kc_state"] = "kc_state"

NAME = "name"
VERSION = 1
PARAMS = {}
ROLE = "single"


class TestSDK(unittest.TestCase):
    """Tests that sdk is supported by current server introspection."""

    def setUp(self):
        """Set mock_server url."""
        set_server_url("http://localhost:4000/")

    def test_list_model(self):
        """Test sdk.list_model method returns correct answer."""
        list_model = sdk.list_model()
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_model), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_model, pandas.core.frame.DataFrame)

    def test_list_dataset_loader(self):
        """Test sdk.list_dataset_loader method returns correct answer."""
        list_dataset_loader = sdk.list_dataset_loader()
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_dataset_loader), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_dataset_loader, pandas.core.frame.DataFrame)

    def test_list_executor(self):
        """Test sdk.list_executor method returns correct answer."""
        list_executor = sdk.list_executor()
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_executor), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_executor, pandas.core.frame.DataFrame)

    def test_add_ml_job(self):
        """Test sdk.add_ml_job method returns correct answer."""
        self.assertIsInstance(
            sdk.add_ml_job(
                job_executor_name=NAME,
                executor_params=PARAMS,
                models_pattern=ArbitraryModels(
                    models=[
                        ModelWithRole(role=ROLE, model_version_choice=ModelVersionChoice(name=NAME, version=VERSION))
                    ]
                ),
                data_pattern=SingleDatasetLoader(
                    name=NAME,
                    version=VERSION,
                    params=[DatasetLoaderMethodParams(params=PARAMS)],
                    collector_params=PARAMS,
                ),
                job_executor_version=VERSION,
            ),
            str,
        )

    def test_job_by_name(self):
        """Test sdk.job_by_name method returns correct answer."""
        job = sdk.job_by_name("job_name")
        self.assertIsInstance(job, ExecutionJob)

    def test_job_metric_by_name(self):
        """Test sdk.job_metric_by_name method returns correct answer."""
        metric = sdk.job_metric_by_name(name=NAME)
        # mock server generates non-empty result by introspection.
        self.assertEqual(len(metric), 1)
        # sdk returns an object of the correct type.
        self.assertIsInstance(sdk.job_metric_by_name(name=NAME), pandas.core.frame.DataFrame)

    def test_list_model_version(self):
        """Test sdk.list_model_version method returns correct answer."""
        list_model_version = sdk.list_model_version(name=NAME)
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_model_version), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_model_version, pandas.core.frame.DataFrame)

    def test_list_dataset_loader_version(self):
        """Test sdk.list_dataset_loader_version method returns correct answer."""
        list_dataset_loader_version = sdk.list_dataset_loader_version(name=NAME)
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_dataset_loader_version), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_dataset_loader_version, pandas.core.frame.DataFrame)

    def test_list_executor_version(self):
        """Test sdk.list_executor_version method returns correct answer."""
        list_executor_version = sdk.list_executor_version(name=NAME)
        # mock server generates non-empty result by introspection.
        self.assertGreater(len(list_executor_version), 0)
        # sdk returns an object of the correct type.
        self.assertIsInstance(list_executor_version, pandas.core.frame.DataFrame)

    def test_model_version_metainfo(self):
        """Test sdk.model_version_metainfo method returns correct answer."""
        self.assertIsInstance(sdk.model_version_metainfo(name=NAME, version=VERSION), ModelVersion)

    def test_rebuild_model_version_image(self):
        """Test sdk.rebuild_model_version_image method returns correct answer."""
        self.assertIsInstance(sdk.rebuild_model_version_image(name=NAME, version=VERSION), str)

    def test_available_metrics(self):
        """Test sdk.available_metrics method returns correct answer."""
        self.assertIsInstance(sdk.available_metrics("job_name"), list)
        self.assertIsInstance(sdk.available_metrics("job_name")[0], str)

    def test_metric_history(self):
        """Test sdk.available_metrics method returns correct answer."""
        response = sdk.metric_history("job_name", "metric")
        self.assertIsInstance(response, tuple)
        self.assertIsInstance(response[0], list)

    def test_get_required_classes_by_executor(self):
        """Test sdk.get_required_classes_by_executor method returns correct answer."""
        required_classes = sdk.get_required_classes_by_executor(name=NAME, version=VERSION)
        self.assertEqual(required_classes, {})

    def test_print_model_schema_for_executor(self):
        """Test a successful sdk.print_model_schema_for_executor call."""
        sdk.print_model_schema_for_executor(
            executor_name=NAME, models=[{"name": NAME, "version": VERSION, "role": ROLE}], executor_version=VERSION
        )

    def test_generate_model_params_for_executor(self):
        """Test sdk.generate_model_params_for_executor method returns correct answer."""
        model_params = sdk.generate_model_params_for_executor(
            executor_name=NAME, models=[{"name": NAME, "version": VERSION, "role": ROLE}], executor_version=VERSION
        )
        self.assertTrue(isinstance(model_params, list) or isinstance(model_params, ArbitraryModels))
        if isinstance(model_params, list):
            for model in model_params:
                self.assertIsInstance(model, ModelMethodParams)
        else:
            self.assertGreaterEqual(len(model_params.models), 1)
            self.assertEqual(model_params.models[0].role, ROLE)
            self.assertEqual(NAME, model_params.models[0].model_version_choice.name)
            self.assertIsInstance(model_params.models[0].params, list)
            self.assertGreaterEqual(len(model_params.models[0].params), 1)
            self.assertIsInstance(model_params.models[0].params[0], ModelMethodParams)

    def test_print_datasetloader_schema(self):
        """Test a successful sdk.print_datasetloader_schema call."""
        sdk.print_datasetloader_schema(name=NAME, version=VERSION)

    def test_print_executor_schema(self):
        """Test a successful sdk.print_executor_schema call."""
        sdk.print_executor_schema(name=NAME, version=VERSION)

    def test_print_executor_roles(self):
        """Test a successful sdk.print_executor_roles call."""
        sdk.print_executor_roles(name=NAME, version=VERSION)

    def test_cancels(self):
        """Test a successful sdk.cancel_* calls."""
        self.assertIsInstance(sdk.cancel_build_job_for_model_version(model_name=NAME, model_version=VERSION), bool)
        self.assertIsInstance(
            sdk.cancel_build_job_for_executor_version(executor_name=NAME, executor_version=VERSION), bool
        )
        self.assertIsInstance(sdk.cancel_venv_build_job_for_model_version(model_name=NAME, model_version=VERSION), bool)
        self.assertIsInstance(sdk.cancel_job("job"), bool)

    def test_delete_objects(self):
        """Test a successful sdk.delete_* calls."""
        self.assertIsInstance(sdk.delete_model_version(model_name=NAME, model_version=VERSION), bool)
        self.assertIsInstance(sdk.delete_model(model_name=NAME), bool)
        self.assertIsInstance(sdk.delete_executor_version(executor_name=NAME, executor_version=VERSION), bool)
        self.assertIsInstance(sdk.delete_executor(executor_name=NAME), bool)
        self.assertIsInstance(sdk.delete_dataset_loader(dataset_loader_name=NAME), bool)
        self.assertIsInstance(
            sdk.delete_dataset_loader_version(dataset_loader_name=NAME, dataset_loader_version=VERSION), bool
        )


if __name__ == "__main__":
    unittest.main()



from ML_management.mlmanagement import mlmanagement

set_experiment = mlmanagement.set_experiment
start_run = mlmanagement.start_run
log_model = mlmanagement.log_model
log_object_src = mlmanagement.log_object_src
log_executor_src = mlmanagement.log_executor_src
log_dataset_loader_src = mlmanagement.log_dataset_loader_src
log_model_src = mlmanagement.log_model_src
log_metric = mlmanagement.log_metric
set_tag = mlmanagement.set_tag
autolog = mlmanagement.autolog
load_model = mlmanagement.load_model
save_model = mlmanagement.save_model
active_run = mlmanagement.active_run
end_run = mlmanagement.end_run
log_artifact = mlmanagement.log_artifact
log_param = mlmanagement.log_param
log_params = mlmanagement.log_params
set_server_url = mlmanagement.set_server_url
get_server_url = mlmanagement.get_server_url
set_s3_url = mlmanagement.set_s3_url
get_s3_gateway_url = mlmanagement.get_s3_gateway_url
set_mlm_credentials = mlmanagement.set_mlm_credentials
get_server_websocket_url = mlmanagement.get_server_websocket_url
MlflowClient = mlmanagement.MlflowClient
